@@ -48,7 +48,10 @@ struct output_buffer_t
       output_format = HB_BUFFER_SERIALIZE_FORMAT_TEXT;
     else
       output_format = hb_buffer_serialize_format_from_string (options.output_format, -1);
-    if (!hb_buffer_serialize_format_to_string (output_format))
+    /* An empty "output_format" parameter basically skips output generating.
+     * Useful for benchmarking. */
+    if ((!options.output_format || *options.output_format) &&
+	!hb_buffer_serialize_format_to_string (output_format))
     {
       if (options.explicit_output_format)
 	fail (false, "Unknown output format `%s'; supported formats are: %s",
